#define _GNU_SOURCE
#include "menu.hh"
#include "graphics.hh"
#include "session.hh"

#include <stdlib.h>


/*

  Resources allocated for each menu
  Window - menu frame, title,
  linked list - menuitems
  GC - titleGC
  dynamic number of menu items

*/

BlackboxMenu::BlackboxMenu(BlackboxSession *ctrl)
{
  session = ctrl;
  display = session->control();
  show_title = True;
  moving = False;
  menu.x = 10;
  menu.y = 10;

  menu.item_pixmap = None;
  menu.pushed_pixmap = None;

  menu.height = menu.title_h = session->titleFont()->ascent +
    session->titleFont()->descent + 8;
  menu.button_h = menu.title_h - 6;
  menu.button_w = menu.button_h * 2 / 3;
  menu.label = "Blackbox";
  menu.width = menu.title_w = XTextWidth(session->titleFont(), menu.label,
			    strlen(menu.label)) + 8;
  menu.item_h = session->menuFont()->ascent +
    session->menuFont()->descent + 4;
  
  unsigned long attrib_mask = CWBackPixmap|CWBackPixel|CWBorderPixel|
    CWOverrideRedirect|CWCursor|CWEventMask;
  XSetWindowAttributes attrib;
  attrib.background_pixmap = None;
  attrib.background_pixel = attrib.border_pixel = session->framePixel();
  attrib.override_redirect = True;
  attrib.cursor = session->sessionCursor();
  attrib.event_mask = StructureNotifyMask|SubstructureNotifyMask|
    SubstructureRedirectMask|ButtonPressMask|ButtonReleaseMask|
    ButtonMotionMask|ExposureMask;

  menu.frame =
    XCreateWindow(display, session->Root(), menu.x, menu.y, menu.title_w,
		  menu.height, 1, session->Depth(), InputOutput,
		  session->visual(), attrib_mask, &attrib);
  XSaveContext(display, menu.frame, session->menuContext(), (XPointer) this);

  attrib_mask = CWBackPixmap|CWBackPixel|CWBorderPixel|CWCursor|CWEventMask;
  attrib.background_pixel = session->framePixel();
  attrib.event_mask |= EnterWindowMask|LeaveWindowMask;
  
  if (show_title) {
    menu.title =
      XCreateWindow(display, menu.frame, 0, 0, menu.title_w, menu.height, 0,
		    session->Depth(), InputOutput, session->visual(),
		    attrib_mask, &attrib);
    XSaveContext(display, menu.title, session->menuContext(),
		 (XPointer) this);
  } else
    menu.title = None;

  menuitems = new llist<BlackboxMenuItem>;
  
  XGCValues gcv;
  gcv.foreground = session->getColor("white");
  gcv.font = session->titleFont()->fid;
  titleGC = XCreateGC(display, menu.frame, GCForeground|GCFont, &gcv);

  itemContext = XUniqueContext();
  //
  // even though this is the end of the constructor the menu is still not
  // completely created.
  //
}



/*

  Resources deallocated for each menu
  Window - menu frame, title,
  linked list - menuitems
  GC - titleGC
  dynamic number of menu items - windows and contexts... then the actual item
      itself is deleted

      the linked list destroys the internal node... not the data contained in
          the node

*/

BlackboxMenu::~BlackboxMenu(void) {
  XUnmapWindow(display, menu.frame);

  int i;
  for (i = 0; i < menuitems->count(); ++i) {
    BlackboxMenuItem *itmp = menuitems->at(i);
    XDeleteContext(display, itmp->window, session->menuContext());
    XDeleteContext(display, itmp->window, itemContext);
    XDestroyWindow(display, itmp->window);
    delete itmp;
  }

  delete menuitems;

  XFreeGC(display, titleGC);
  if (show_title) {
    XDeleteContext(display, menu.title, session->menuContext());
    XDestroyWindow(display, menu.title);
  }
  XDeleteContext(display, menu.frame, session->menuContext());
  XDestroyWindow(display, menu.frame);
}


int BlackboxMenu::insert(char **label) {
  BlackboxMenuItem *item = new BlackboxMenuItem(createItemWindow(), label);
  XSaveContext(display, item->window, itemContext, (XPointer) item);
  menuitems->append(item);

  return menuitems->count();
}


int BlackboxMenu::insert(char *label) {
  BlackboxMenuItem *item = new BlackboxMenuItem(createItemWindow(), label);
  XSaveContext(display, item->window, itemContext, (XPointer) item);
  menuitems->append(item);
  
  return menuitems->count();
}


int BlackboxMenu::insert(char *label, int function, char *exec) {
  int ret = 0;

  switch (function) {
  case BlackboxSession::B_Execute: {
    BlackboxMenuItem *item = new BlackboxMenuItem(createItemWindow(), label,
					       exec);
    XSaveContext(display, item->window, itemContext, (XPointer) item);
    menuitems->append(item);

    ret = menuitems->count();
    break; }

  case BlackboxSession::B_Exit:
  case BlackboxSession::B_Restart: {
    BlackboxMenuItem *item = new BlackboxMenuItem(createItemWindow(), label,
						  function);
    XSaveContext(display, item->window, itemContext, (XPointer) item);
    menuitems->append(item);

    ret = menuitems->count();
    break; }

  case BlackboxSession::B_Shutdown:
    //    ret = insert(label, (void (*)()) session->Shutdown);
    break;

  case BlackboxSession::B_WindowClose:
  case BlackboxSession::B_WindowIconify:
  case BlackboxSession::B_WindowRaise:
  case BlackboxSession::B_WindowLower:
    if (! show_title) {
      BlackboxMenuItem *item =
	new BlackboxMenuItem(createItemWindow(), label, function);
      XSaveContext(display, item->window, itemContext, (XPointer) item);
      menuitems->append(item);
      
      ret = menuitems->count();
    }
    break;
  }
  
  return ret;
}


int BlackboxMenu::insert(char *label, BlackboxMenu *submenu) {
  BlackboxMenuItem *item = new BlackboxMenuItem(createItemWindow(), label,
						submenu);
  menuitems->append(item);

  return menuitems->count();
}


int BlackboxMenu::remove(char *) {
  // i hope i never need this function
  return 0;
}


int BlackboxMenu::remove(int index) {
  BlackboxMenuItem *item = menuitems->at(index);
  menuitems->remove(index);

  XDeleteContext(display, item->window, itemContext);
  XDestroyWindow(display, item->window);
  delete item;

  return menuitems->count();
}


void BlackboxMenu::updateMenu(void) {
  int i = 0, ii = 0;
  BlackboxMenuItem *itmp = NULL;

  menu.width =  XTextWidth(session->titleFont(), menu.label,
                           strlen(menu.label)) + 8;
  for (i = 0; i < menuitems->count(); ++i) {
    itmp = menuitems->at(i);
    ii = XTextWidth(session->menuFont(),
                    ((itmp->ulabel) ? *itmp->ulabel : itmp->label),
                    strlen(((itmp->ulabel) ? *itmp->ulabel : itmp->label))) + 8;
    menu.width = ((menu.width < (unsigned int) ii) ? ii : menu.width);
  }

  if (show_title) {
    BImage mt_image(menu.width, menu.title_h, session->Depth(),
		    session->menuColor());
    mt_image.renderBevel();
    Pixmap mt_pixmap = mt_image.convertToPixmap(display, session->visual(),
						session->Root());
    XSetWindowBackgroundPixmap(display, menu.title, mt_pixmap);
    XClearWindow(display, menu.title);
    XFreePixmap(display, mt_pixmap);
    XDrawString(display, menu.title, titleGC, 3, session->titleFont()->ascent
		+ 3, menu.label , strlen(menu.label));
  }

  BImage mi_image(menu.width, menu.item_h, session->Depth(),
		  session->menuItemColor());
  mi_image.renderButton();

  if (menu.item_pixmap) XFreePixmap(display, menu.item_pixmap);
  menu.item_pixmap = mi_image.convertToPixmap(display, session->visual(),
					      session->Root());

  XGrabServer(display);
  menu.height = ((show_title) ? menu.title_h : 0 ) +
    ((menu.item_h + 1) * menuitems->count());
  XResizeWindow(display, menu.frame, menu.width, menu.height);
  if (show_title) XResizeWindow(display, menu.title, menu.width, menu.title_h);

  for (i = 0; i < menuitems->count(); ++i) {
    itmp = menuitems->at(i);
    XSetWindowBackgroundPixmap(display, itmp->window, menu.item_pixmap);
    XMoveResizeWindow(display, itmp->window, 0, (i * menu.item_h) +
		      ((show_title) ? menu.title_h : 0) + i + 1,
		      menu.width + 8, menu.item_h);
    XClearWindow(display, itmp->window);
    XDrawString(display, itmp->window, titleGC, 4,
		session->titleFont()->ascent + 2,
		((itmp->ulabel) ? *itmp->ulabel : itmp->label),
		strlen(((itmp->ulabel) ? *itmp->ulabel : itmp->label)));
  }

  if (menu.pushed_pixmap) XFreePixmap(display, menu.pushed_pixmap);
  BImage pushed_image(menu.width, menu.item_h, session->Depth(),
		      session->menuItemToColor());
  pushed_image.renderButton();
  pushed_image.invertImage();
  menu.pushed_pixmap =
    pushed_image.convertToPixmap(display, session->visual(),
				 session->Root());
  
  XMapSubwindows(display, menu.frame);
  XFlush(display);
  XSync(display, 0);
  XUngrabServer(display);
}


void BlackboxMenu::showMenu(void) {
  XGrabServer(display);
  XMapSubwindows(display, menu.frame);
  XMapRaised(display, menu.frame);
  updateMenu();
  XUngrabServer(display);
}


void BlackboxMenu::buttonPressEvent(XButtonEvent *be) {
  if (show_title && be->window == menu.title) {
    if (be->button == 1) XRaiseWindow(display, menu.frame);
    else if (be->button == 2) XLowerWindow(display, menu.frame);
    titlePressed(be->button);
  } else {
    XSetWindowBackgroundPixmap(display, be->window, menu.pushed_pixmap);
    XClearWindow(display, be->window);
    int i;

    for (i = 0; i < menuitems->count(); ++i) {
      BlackboxMenuItem *item = menuitems->at(i);
      if (item->window == be->window) {
	XDrawString(display, item->window, titleGC, 4,
		    session->titleFont()->ascent + 2,
		    ((item->ulabel) ? *item->ulabel : item->label),
		    strlen(((item->ulabel) ? *item->ulabel : item->label)));
	
	itemPressed(be->button, i);
	break;
      }
    }
  }
}


void BlackboxMenu::buttonReleaseEvent(XButtonEvent *re) {
  if (show_title && re->window == menu.title) {
    if (moving) {
      XUngrabPointer(display, CurrentTime);
      moving = False;
    }
    titleReleased(re->button);
  } else {
    XSetWindowBackgroundPixmap(display, re->window, menu.item_pixmap);
    XClearWindow(display, re->window);
    int i;

    for (i = 0; i < menuitems->count(); ++i) {
      BlackboxMenuItem *item = menuitems->at(i);
      if (item->window == re->window) {
	XDrawString(display, item->window, titleGC, 4,
		    session->titleFont()->ascent + 2,
		    ((item->ulabel) ? *item->ulabel : item->label),
		    strlen(((item->ulabel) ? *item->ulabel : item->label)));

	if (re->x > 0 && re->x < (int) menu.width &&
	    re->y > 0 && re->y < (int) menu.item_h)
	  itemReleased(re->button, i);
	
	break;
	  
	/*	  if (re->button == 1) {
	    if (window) {

	    } else if (ws_manager) {
	      if (i != ws_manager->currentWorkspaceID())
		ws_manager->changeWorkspaceID(i);
	    } else if (item->callback()) {
	      item->callback()();
	    } else
	    
	    break;
	  }
	}
      }
    }
  }
  }*/
      }
    }
  }
}

void BlackboxMenu::motionNotifyEvent(XMotionEvent *me) {
  if (me->window == menu.title && session->button1Pressed()) {
    if (! moving) {
      if (XGrabPointer(display, menu.title, False, PointerMotionMask|
		       ButtonReleaseMask, GrabModeAsync, GrabModeAsync,
		       None, session->moveCursor(), CurrentTime)
	  == GrabSuccess) {
	moving = 1;
	menu.x_move = me->x;
	menu.y_move = me->y;
      } else
 	moving = False;	
    } else {
      int dx = me->x_root - menu.x_move,
	dy = me->y_root - menu.y_move;
      XMoveWindow(display, menu.frame, dx, dy);
    }
  }
}


void BlackboxMenu::exposeEvent(XExposeEvent *ee) {
  if (ee->window == menu.title) {
    XDrawString(display, menu.title, titleGC, 3, session->titleFont()->ascent
		+ 3, menu.label, strlen(menu.label));
  } else {
    int i;

    for (i = 0; i < menuitems->count(); ++i) {
      BlackboxMenuItem *item = menuitems->at(i);
      if (item->window == ee->window) {
	XDrawString(display, item->window, titleGC, 4,
		    session->titleFont()->ascent + 2,
		    ((item->ulabel) ? *item->ulabel : item->label),
		    strlen(((item->ulabel) ? *item->ulabel : item->label)));

	break;
      }
    }
  }
}


Window BlackboxMenu::createItemWindow(void) {
  unsigned long attrib_mask = CWBackPixmap|CWBackPixel|CWBorderPixel|CWCursor|
    CWEventMask;
  
  XSetWindowAttributes attrib;
  attrib.background_pixmap = None;
  attrib.background_pixel = session->framePixel();
  attrib.border_pixel = None;
  attrib.override_redirect = True;
  attrib.cursor = session->sessionCursor();
  attrib.event_mask = StructureNotifyMask|SubstructureNotifyMask|
    SubstructureRedirectMask|ButtonPressMask|ButtonReleaseMask|
    ButtonMotionMask|ExposureMask|EnterWindowMask|LeaveWindowMask;
  
  Window c =
    XCreateWindow(display, menu.frame, 0, (menu.item_h * menuitems->count())
		  + menu.title_h + 1, 1, menu.item_h, 0, session->Depth(),
		  InputOutput, session->visual(), attrib_mask, &attrib);

  XSaveContext(display, c, session->menuContext(), (XPointer) this);
  return c;
}


void BlackboxMenu::hideMenu(void) {
  XGrabServer(display);
  XUnmapWindow(display, menu.frame);
  XUngrabServer(display);
}


void BlackboxMenu::moveMenu(int x, int y) {
  XGrabServer(display);
  XMoveWindow(display, menu.frame, x, y);
  XUngrabServer(display);
}


Window BlackboxMenu::windowID(void) { return menu.frame; }
void BlackboxMenu::setMenuLabel(char *n) { menu.label = n; }
void BlackboxMenu::showTitle(void) { show_title = True; }
void BlackboxMenu::hideTitle(void) { show_title = False; }
BlackboxMenuItem *BlackboxMenu::at(int i) { return menuitems->at(i); }
BlackboxSession *BlackboxMenu::Session(void) { return session; }
